//////////////////////////////////////////////////////////////////////////////
//
// CG_Steihaug_minimizer.cpp, 10/25/02, Ilja Friedel 
//
//
//////////////////////////////////////////////////////////////////////////////

#include "Minimizer.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
CG_Steihaug_Minimizer::~CG_Steihaug_Minimizer()
{
  delete[] X;
  delete[] p;
  delete[] p1;
  delete[] p2;
  delete[] r;
  delete[] g;
  delete[] d;
  delete[] Bd;
  delete[] Bp;
}

//////////////////////////////////////////////////////////////////////////////

CG_Steihaug_Minimizer::CG_Steihaug_Minimizer(double Db, double et)
{
  Delta_bar=Db;
  eta=et;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Norcedal/Wright p68
//
// where does eta come from? Forcing sequence? No! prediction accuracy

void CG_Steihaug_Minimizer::minimize(Mesh *M)
{
  double f, f_new, m_new;
  double *X_new;
  double Delta, eps;
  double rho;
  double precision, precision_old;
  int    boundary_flag;
  int    iteration, max_iteration;
  int    Newton_converged;

  assert(Delta_bar>0.0); // should Delta_bar==1? *****
  assert(eta>=0.0 && eta < 0.25);
  assert(M);
  mesh=M;
  numpoints=M->numpoints;
  assert(numpoints>0);
  assert(numpoints<=100000);

  X    = new double[3*numpoints];  // current variables/points
  p    = new double[3*numpoints];  // descent direction
  p1   = new double[3*numpoints];  // temp
  p2   = new double[3*numpoints];  // temp
  r    = new double[3*numpoints];  // residual
  g    = new double[3*numpoints];  // gradient
  d    = new double[3*numpoints];  // 
  Bd   = new double[3*numpoints];  // temp
  Bp   = new double[3*numpoints];  // temp
  X_new= new double[3*numpoints];  // temp

  set(p,0.0);
  copy(mesh->P,X);       // initialize point positions
  f = Energy(X);         // initialize energy
  //m = ModelEnergy(X,p);  // initialize second order approximation
  Delta = 1.0;          // some guess for initial trust region radius
  rho = 1.0;

  gradEnergy(X,g);
  B=HessEnergy(X);
  
  iteration        = 0;
  max_iteration    = 1000;
  Newton_converged = 0;
  precision        = 0.0;
  precision_old    = 0.0;
  f_new            = 2.0 * f;

  while(true)
    {
      // page 138, forcing quadratic convergence ????? ******
      eps=pow(dot(g,g),0.5);
      printf("\noeps=%1.6e  ",eps);
      if(eps>0.5)      eps=0.5;
      if(eps<EPSILON2) eps=EPSILON2;

      printf("Energy=%1.6e  eps=%1.6e  Delta=%1.6e  rho=%1.6e",f,eps,Delta,rho);
      // accuracy of subsolution
      
      // disappearing trust ...
      if(Delta<=EPSILON1)
	{
	  cerr << "Warning: trust region radius tiny!" << endl;
	  copy(X,mesh->P);
	  return;
	}
      
      // find our descent direction
      boundary_flag=CG_Steihaug_direction(B,g,X,p,Delta,eps);      

      // what really does this tell us? ***********
      if(boundary_flag==-1 && eps<1e-6)
	{
	  copy(X,mesh->P);
	  return; // end of minimization
	}
      //      assert(sqrt(dot(p,p))<Delta+EPSILON2);

      add(X,p,X_new);
      // evaluate rho (p67)
      f_new = Energy(X_new);    // energy
      m_new = ModelEnergy(B,g,X,p); // prediction of the second order model

//////////////////////////////////////////////////////////////////////////////
      // keeping track of Newton convergence
      if(fabs(f-f_new)<EPSILON1) precision=17;
      else
	{
	  // observed negative precision! ********
	  if(fabs(f_new)>EPSILON1) precision=-log10((fabs((f-f_new)/f_new)));  
	  else                     precision=-log10(EPSILON1);
	}

      cout << "  precision=" << precision << endl;

      if(Delta>=0.75 && precision-precision_old>=0.5)
	{
	      Newton_converged++;
	}
      else
	{
	  if(Newton_converged>=3) cerr << "Newton convergence aborted." << endl;
	  Newton_converged=0;
	  max_iteration=1000;
	}
      precision_old=precision;

      // test for Newton convergence
      if(Newton_converged==3)
	{
	  cerr << "Newton convergence reached." << endl;
	  if(Newton_converged==3) max_iteration=iteration+10;
	}

//////////////////////////////////////////////////////////////////////////////
      
      if(f-m_new>0.0)
	{
	  rho = (f-f_new)/(f-m_new); // how well did the model perform?
	}
      else
	{
	  rho=EPSILON1;
	}

//////////////////////////////////////////////////////////////////////////////

      if(rho<0.25) // bad performance of model, reduce trust radius
	{
	  cout << "Delta before="<< Delta<< endl;
	  Delta=0.25*sqrt(dot(p,p));
	  cout << "Delta after="<< Delta<< endl;
	}
      else
	{
	  if(rho>0.75 && boundary_flag==1) // good performance and boundary hit
	    {
	      //compute Delta=min(2.0*Delta,Delta_bar)
	      Delta=2.0*Delta;
	      if(Delta>Delta_bar) Delta=Delta_bar;
	    }
	  // else do nothing
	}
      if(rho>eta) // accept X_new if model was reasonably good
	{
	  copy(X_new,X);
	  f = f_new;
	  gradEnergy(X,g);          // gradient
	  delete B;
	  B=HessEnergy(X);
	}
      // else do nothing

//////////////////////////////////////////////////////////////////////////////

      if(!(iteration%20))//&& iteration>0)
	{
	  char fname[256];
	  sprintf(fname,"gmin%d.iv",iteration);
	  mesh->export_IV(fname,X,g);
	  set(Bp,0.0);
	  sprintf(fname,"min%d.iv",iteration);
	  mesh->export_IV(fname,X,Bp);
	  //sprintf(fname,"min%d",iteration);
	  //mesh->save(fname);
	}
      /*
      if(iteration>60)//&& iteration>0)	
	{
	  char fname[256];
	  sprintf(fname,"gmin%d.iv",iteration);
	  mesh->export_IV(fname,X,g);
	}
      */
      if(iteration==max_iteration)
	{
	  copy(X,mesh->P);
	  return;
	}
      iteration++;
    }
}

//////////////////////////////////////////////////////////////////////////////
// Norcedal/Wright p75

int CG_Steihaug_Minimizer::CG_Steihaug_direction(DynamicCompCol_Mat_double *B, 
						 double *grad,
						 double *x_old, 
						 double *p, 
						 double Delta, double eps)
{
  double alpha, beta;
  double rr, rr0, rnrn, pnpn;
  double dBd;

  assert(eps  >0.0);
  assert(Delta>0.0);

  //  set(p,0.0);   // p=0
  copy(grad,r); // r=g
  ncopy(r,d);   // d=-r

  rr0=dot(r,r);
  if(rr0<eps*eps)
    {
      cerr << "Warning: should return." << endl;
      //return(-1); // p=0 and already have desired accuracy *************
    }
  set(p,0.0);   // switched from above
  rr=rr0;

  while(true)
    {
      B->mult(d,Bd); // Bd=B*d
      dBd=dot(d,Bd); // d^T*B*T
      if(dBd <= 0.0) // we encounter direction of negative curvature
	{
	  cout << "[-]";
	  // find tau such that
	  //    p = pj + tau dj 
	  // minimizes 
	  //    m(p)=f + g^T*p + 0.5*p^T*B*p  
	  // and p satisfies
	  //    ||p||=Delta
	  //
	  // check both roots of equation

	  double pp,pd,dd,tau1,tau2;
	  double f1, f2;

	  pp=dot(p,p);
	  pd=dot(p,d);
	  dd=dot(d,d);

	  // C(simplify(solve(dd*tau*tau+2*pd*tau+pp-Delta*Delta,tau)[1]));
	  tau1=(-pd+sqrt(pd*pd-dd*pp+dd*Delta*Delta))/dd;
	  // 2. solution quadratic equation
	  tau2=-(pd+sqrt(pd*pd-dd*pp+dd*Delta*Delta))/dd;

	  scale_add(p,tau1,d,p1);
	  scale_add(p,tau2,d,p2);

	  // why not
	  // f1 = Energy(x+p1);
	  // f2 = Energy(x+p2);
	  // much cheaper! (and accurate)
	  
	  f1 = ModelEnergy(B,grad,x_old,p1);
	  f2 = ModelEnergy(B,grad,x_old,p2);
	  
	  if(f1<f2) copy(p1,p);
	  else      copy(p2,p);

	  return(1); // we hit the boundary
	}

      alpha=rr/dBd;  //cout << "alpha=" << alpha << "   ";

      scale_add(p,alpha,d,p); // p_i is now p_i+1

      pnpn=dot(p,p); //cout << "pnpn=" << pnpn << "  ";
      if(pnpn>=Delta*Delta) // we are leaving the trust region
	{
	  cout << "[B]";
	  // Find tau>=0 such that
	  //    p=pj + tau*dj
	  // and p satisfies
	  //    ||p||=Delta
	  // 
	  // only consider positive root

	  double pp,pd,dd,tau1,tau2;

	  pp=dot(p,p);
	  pd=dot(p,d);
	  dd=dot(d,d);

	  tau1=(-pd+sqrt(pd*pd-dd*pp+dd*Delta*Delta))/dd;
	  tau2=-(pd+sqrt(pd*pd-dd*pp+dd*Delta*Delta))/dd;

	  if(tau1+alpha>=0.0)      scale_add(p,tau1,d,p);
	  else if(tau2+alpha>=0.0) scale_add(p,tau2,d,p);
	  else
	    {
	      cerr << "Error: tau1 & tau2 < 0.0!" << endl;
	      exit(1);
	    }

	  return(1); // we hit the boundary
	}
      
      scale_add(r,alpha,Bd,r); // r_i is now r_i+1
      rnrn=dot(r,r); //cout << "rnrn=" << rnrn << "   ";
      if(rnrn<eps*eps*rr0) // reached desired accuracy of solution
	{
	  cout << "[A]";
	  // just return p_new
	  return(0); // no boundary hit
	}

      beta=rnrn/rr;
      scale_add(-1.0,r,beta,d,d); // ***** changed from book, see page 111 *****
      rr=rnrn;

      // NEVER change x_old!
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
